/*************************************************************************/
/* Copyright (C) 2007,2008 sacamano <m.sacamano@gmail.com>		 */
/* 									 */
/* This program is free software: you can redistribute it and/or modify	 */
/* it under the terms of the GNU General Public License as published by	 */
/* the Free Software Foundation, either version 3 of the License, or	 */
/* (at your option) any later version.					 */
/* 									 */
/* This program is distributed in the hope that it will be useful,	 */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of	 */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the	 */
/* GNU General Public License for more details.				 */
/* 									 */
/* You should have received a copy of the GNU General Public License	 */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*************************************************************************/

#include "consonance.h"

static snd_mixer_t *alsa_mixer = NULL;
static snd_mixer_elem_t *alsa_elem = NULL;
static GIOChannel **chans = NULL;
static int n_chans = 0;

static int alsa_elem_cb(snd_mixer_elem_t *alsa_elem, unsigned int mask)
{
	glong curr_vol;
	struct con_win *cwin;

	cwin = snd_mixer_elem_get_callback_private(alsa_elem);
	snd_mixer_selem_get_playback_volume(alsa_elem,
					    SND_MIXER_SCHN_FRONT_LEFT,
					    &curr_vol);
	cwin->cmixer->curr_vol = SCALE_UP_VOL(curr_vol);
	gtk_scale_button_set_value(GTK_SCALE_BUTTON(cwin->vol_button), cwin->cmixer->curr_vol);

	return 0;
}

static gboolean alsa_io_cb(GIOChannel *chan, GIOCondition condition, gpointer data)
{
	switch(condition) {
	case G_IO_OUT:
	case G_IO_IN:
	case G_IO_PRI:
		snd_mixer_handle_events(alsa_mixer);
		return TRUE;
	case G_IO_ERR:
	case G_IO_HUP:
	case G_IO_NVAL:
	default:
		g_warning("(%s): ALSA mixer polling error\n", __func__);
		return FALSE;
	}
}

gint alsa_init_mixer(struct con_win *cwin)
{
#define POLL_FLAGS G_IO_IN | G_IO_OUT | G_IO_PRI | G_IO_ERR | G_IO_HUP | G_IO_NVAL
	gint ret = 0, cnt = 0, i;
	snd_mixer_elem_t *elem = NULL;
	const gchar *elem_name = NULL;
	struct pollfd *fds;

	if (snd_mixer_open(&alsa_mixer, 0) < 0) {
		g_critical("(%s): Unable to open Alsa mixer", __func__);
		return -1;
	}

	if (snd_mixer_attach(alsa_mixer, ALSA_MIXER_DEV) < 0) {
		g_critical("(%s): Unable to attach to alsa dev : %s", __func__, ALSA_MIXER_DEV);
		ret = -1;
		goto bad;
	}

	if (snd_mixer_selem_register(alsa_mixer, NULL, NULL) < 0) {
		g_critical("(%s): Unable to register alsa mixer simple element", __func__);
		ret = -1;
		goto bad;
	}

	if (snd_mixer_load(alsa_mixer) < 0) {
		g_critical("(%s): Unable to load mixer", __func__);
		ret = -1;
		goto bad;
	}

	elem = snd_mixer_first_elem(alsa_mixer);

	while (elem) {
		if (snd_mixer_elem_get_type(elem) == SND_MIXER_ELEM_SIMPLE) {
			elem_name = snd_mixer_selem_get_name(elem);
			if (!g_ascii_strcasecmp(elem_name, ALSA_MIXER_ELEM)) {
				CDEBUG("(%s): Found ALSA elem: %s", __func__, ALSA_MIXER_ELEM);
				break;
			}
		}
		elem = snd_mixer_elem_next(elem);
	}
	alsa_elem = elem;
	if (!alsa_elem) {
		g_critical("(%s): Unable to find alsa mixer element : %s",
			   __func__, ALSA_MIXER_ELEM);
		ret = -1;
		goto bad;
	}

	snd_mixer_selem_get_playback_volume_range(alsa_elem,
						  &cwin->cmixer->min_vol,
						  &cwin->cmixer->max_vol);
	snd_mixer_selem_get_playback_volume(alsa_elem,
					    SND_MIXER_SCHN_FRONT_LEFT,
					    &cwin->cmixer->curr_vol);
	CDEBUG("(%s): Max vol: %ld, Curr vol: %ld",
	       __func__, cwin->cmixer->max_vol, cwin->cmixer->curr_vol);

	cnt = snd_mixer_poll_descriptors_count(alsa_mixer);
	if (!cnt) {
		g_warning("(%s): No ALSA poll descriptors\n", __func__);
		goto exit;
	}

	fds = g_new0(struct pollfd, cnt);

	if (snd_mixer_poll_descriptors(alsa_mixer, fds, cnt) < 0) {
		g_warning("(%s): Unable to get ALSA poll descriptors\n", __func__);
		g_free(fds);
		goto exit;
	}

	snd_mixer_elem_set_callback(alsa_elem, alsa_elem_cb);
	snd_mixer_elem_set_callback_private(alsa_elem, cwin);

	chans = g_new0(GIOChannel*, cnt);

	for (i=0; i<cnt; i++) {
		chans[i] = g_io_channel_unix_new(fds[i].fd);
		g_io_add_watch(chans[i], POLL_FLAGS, alsa_io_cb, cwin);
	}
	g_free(fds);
	n_chans = cnt;

	CDEBUG("(%s): Poll ALSA for %d descriptors", __func__, cnt);

	goto exit;
bad:
	snd_mixer_close(alsa_mixer);
exit:
	return ret;
#undef POLL_FLAGS
}

gint alsa_mute_mixer(struct con_win *cwin)
{
	return 0;
}

void alsa_set_volume(struct con_win *cwin)
{
	CDEBUG("(%s): Setting ALSA volume to %ld",
	       __func__, SCALE_DOWN_VOL(cwin->cmixer->curr_vol));

	if (snd_mixer_selem_set_playback_volume_all(alsa_elem,
						    SCALE_DOWN_VOL(cwin->cmixer->curr_vol)) < 0)
		g_critical("(%s): Unable to set alsa volume : %ld",
			   __func__, SCALE_DOWN_VOL(cwin->cmixer->curr_vol));
}

void alsa_inc_volume(struct con_win *cwin)
{
	if (SCALE_DOWN_VOL(cwin->cmixer->curr_vol) < cwin->cmixer->max_vol)
		cwin->cmixer->curr_vol++;

	CDEBUG("(%s): Setting ALSA volume to %ld",
	       __func__, SCALE_DOWN_VOL(cwin->cmixer->curr_vol));

	if (snd_mixer_selem_set_playback_volume_all(alsa_elem,
						    SCALE_DOWN_VOL(cwin->cmixer->curr_vol)) < 0)
		g_critical("(%s): Unable to set alsa volume : %ld",
			   __func__, SCALE_DOWN_VOL(cwin->cmixer->curr_vol));
}

void alsa_dec_volume(struct con_win *cwin)
{
	if (SCALE_DOWN_VOL(cwin->cmixer->curr_vol) > cwin->cmixer->min_vol)
		cwin->cmixer->curr_vol--;

	CDEBUG("(%s): Setting ALSA volume to %ld",
	       __func__, SCALE_DOWN_VOL(cwin->cmixer->curr_vol));

	if (snd_mixer_selem_set_playback_volume_all(alsa_elem,
						    SCALE_DOWN_VOL(cwin->cmixer->curr_vol)) < 0)
		g_critical("(%s): Unable to set alsa volume : %ld",
			   __func__, SCALE_DOWN_VOL(cwin->cmixer->curr_vol));
}

void alsa_deinit_mixer(struct con_win *cwin)
{
	GError *err = NULL;
	int i;

	for (i=0; i<n_chans; i++) {
		g_io_channel_shutdown(chans[i], FALSE, &err);
		if (err) {
			g_warning("(%s): Unable to shutdown IO Channel for ALSA\n", __func__);
			g_error_free(err);
			err = NULL;
		}
		else {
			CDEBUG("(%s): Shutting down ALSA IO channel", __func__);
		}
		g_io_channel_unref(chans[i]);
	}
	g_free(chans);

	if (snd_mixer_detach(alsa_mixer, ALSA_MIXER_DEV))
		g_warning("(%s): Unable to deinit alsa mixer", __func__);

	snd_mixer_free(alsa_mixer);
}
